;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Effect Name : random15
;
; Description : 24-bits random number generator generator
;
; Nb In       : 0
; Nb Out      : 1
;
; Input       : None
; Output      : 24-bits random number between 0 and 1
;
; Cycles      : 79 + 12 cycles (-1=unknownw)
; Memory      : 4 words (without init)
; Delay Line  : no
;
; Author      : Gregor
; Date        : 05/11/2004
; Version     : 1.0
;
; This is implemented using the usual recursive relation
;            x_n = (a x_n-1 + c) mod 2^24
; Note that the chosen values of a and c ensure a maximal periodicity
;   and passes rather successfully a 2-dimensional lattice test
; Note also that due to the way AL3101 deals with multiplications,
;   one may not perform trivially the multiplication by a.
;   Here, we proceed by splitting x, a and c in 4 bunches of 6 bits
;   partial multiplicatoins are made using integer multiplication 
;   (MLTB). For this reason, all bunches are pre-aligned with
;   vanishing 12 least-significant bits.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

ABS	INIT	0x3FF
	
	;; SEC1 Control parameters
ABS	Value1	0x0		; \ Initial value (seed)
ABS	Value2	0x1		; | 30865 as multiplicative constant
ABS	Value3	0x2		; | split into 4 6-bits INT bunches
ABS	Value4	0x3             ; /
	
	;; SEC2 Initial values of control parameters (S3.24 format)
IValue1	EQU	0x0000000	; \ Initial value (seed)
IValue2	EQU	0x0007000	; | 30865 as multiplicative constant
IValue3	EQU	0x0022000	; | split into 4 6-bits INT bunches
IValue4	EQU	0x0011000	; / 
	
	;; SEC3 Constants definitions
CMult1	EQU	0x00280		; \
CMult2	EQU	0x000c0		; | 2634357 as multiplicative constant
CMult3	EQU	0x00240		; | split into 4 6-bits INT bunches in S3.18
CMult4	EQU	0x00d40		; /
CAdd1	EQU	0x0030000	; \ 
CAdd2	EQU	0x0024000	; | 12732917 as additive constant
CAdd3	EQU	0x0027000	; | split into 4 6-bits INT bunches in S3.24
CAdd4	EQU	0x0035000	; /
CMaskLow  EQU	0x003f000	; keep only relevant bits
CMaskHigh EQU	0x0fc0000	; keep only relevant bits
	
	;; SEC4 Memory variables definition

	;; SEC5 Initial values of memory variables

	;; SEC6 Initialization
CM	0x40000	INIT
SKIP	!Z	EndInit
C	IValue1
SCA	0x0 Value1
C	IValue2
SCA	0x0 Value2
C	IValue3
SCA	0x0 Value3
C	IValue4
SCA	0x0 Value4
C       0x1000000		; 1->A
SCA     0x0     INIT		; do not reinitialize next time
EndInit:

	;; SEC7 Module code
	;; step 1: multiplication
	;;   1.1 part 4
DAC	0x0 CMult4		; load 'a4'
XCM	0x40000 Value4		; store it in B and get previous value
MLTB				; 4th part mult
1AC	CAdd4			; add constant
SCA	0x40000 Dir0		; 
ANDC	CMaskLow		; keep relevant bits for new x4
SCA	0x0 Dir4		; save new value of x4
CM 	0x1000 Dir0
ANDC	CMaskLow		; get carry (note U is 0)

	;;   1.2  part 3
SAUC	CMult3 Dir0		; load 'a3'
XCM	0x40000 Value4		; store it in B and get previous value
MLTB				
CMA	0x40000 Dir0
SCA	0x0 Dir0		; adds a3.x4 to carry
DAC	0x0 CMult4
XCM	0x40000 Value3
MLTB
1AC	CAdd3			; add constant
CMA	0x40000	Dir0		
SCA	0x40000 Dir0		; adds a4.x3 to carry
ANDC	CMaskLow		; keep relevant bits for new x3
SCA	0x0 Dir3		; save new value of x3
CM 	0x1000 Dir0
ANDC	CMaskLow		; get carry (note U is 0)
	
	;;   1.3  part 2
SAUC	CMult2 Dir0		; load 'a2'
XCM	0x40000 Value4		; store it in B and get previous value
MLTB				
CMA	0x40000 Dir0
SCA	0x0 Dir0		; adds a2.x4 to carry
DAC	0x0 CMult3
XCM	0x40000 Value3
MLTB
CMA	0x40000	Dir0
SCA	0x0 Dir0		; adds a3.x3 to carry
DAC	0x0 CMult4
XCM	0x40000 Value2
MLTB
1AC	CAdd2			; add constant
CMA	0x40000	Dir0
SCA	0x40000 Dir0		; adds a4.x2 to carry
ANDC	CMaskLow		; keep relevant bits for new x2
SCA	0x0 Dir2		; save new value of x2
CM 	0x1000 Dir0
ANDC	CMaskLow		; get carry (note U is 0)
	
	;;   1.4  part 1
SAUC	CMult1 Dir0		; load 'a1'
XCM	0x40000 Value4		; store it in B and get previous value
MLTB				
CMA	0x40000 Dir0
SCA	0x0 Dir0		; adds a1.x4 to carry
DAC	0x0 CMult2
XCM	0x40000 Value3
MLTB
CMA	0x40000	Dir0
SCA	0x0 Dir0		; adds a2.x3 to carry
DAC	0x0 CMult3
XCM	0x40000 Value2
MLTB
CMA	0x40000	Dir0
SCA	0x0 Dir0		; adds a3.x2 to carry
DAC	0x0 CMult4
XCM	0x40000 Value1
MLTB
1AC	CAdd1			; add constant
CMA	0x40000	Dir0		; adds a4.x1 to carry (no need to keep new carry)
ANDC	CMaskLow		; keep relevant bits for new x2
SCA	0x40000 Value1		; save new value of x1

	;; step 2: save new value(s) and build return value
CAD	0x100000 0x0		; start dealing with bits 1-6 (highest)
CAD	0x100000 0x0
CAD	0x100000 0x0		; 6-bit shift
XCM	0x40000 Dir2		; start dealing with bits 7-12
SCA	0x40000 Value2		; save for next run
CAB	0x40000			; bits 1-12 ok
XCM	0x40000 Dir3		; start dealing with bits 13-18
SCA	0x1000 Value3		; save for next run and perform right shift
CAB	0x40000			; bits 1-18 ok
XCM	0x40000 Dir4		; start dealing with bits 19-24
SCA	0x40 Value4		; save for next run and perform right shift
CAB	0x40000			; bits 1-24 ok
SCA	0x0 OUT1
