;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Effect Name : sinus
;
; Description : Sinusoidal wave generator
;
; Nb In       : 0
; Nb Out      : 1
;
; Input       : nothing
; Output      : Sinusoidal signal
;
; Cycles      : 19 + 12 cycles (-1=unknownw)
; Memory      : 5 words (without init)
; Delay Line  : no
;
; Author      : Gregor
; Date        : 26/02/2003
; Version     : 1.0
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; The sinusoid is generated by generating a triangle and then
;;    applying a polynome of the following form to it:
;;    out = Coeff1 * in + Coeff3 * in^3 + Coeff5 * in^5
;;
;; The approximated sinus generates a contribution at 3*Fc and 5*Fc
;;    for 440Hz operating frequency and 48000 sampling frequency,
;;    the magnitude of the contributions are resp. -96dB and -85dB.
;;
;; Important remark: the amplitude parameter should be set to 0.5
;;    for full-scale operation when the output is directly
;;    connected to a DAC as the DAC is expecting 0.5 as full-scale.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


ABS	INIT	0x3FF
	
	;; SEC1 Control parameters
ABS	Ampl	0x0		; amplitude
ABS	Delta	0x1		; Delta = 4/(smpfreq * T) = 4 * Fc / Fs
	
	;; SEC2 Initial values of control parameters (S3.24 format)
IAmpl	EQU	0.5
IDelta	EQU	0x000a377f	; Delta = 4 * Fc / Fs
						; ex:	 0x00005f2 for 1 Hz
						;        0x00a377f for 440 Hz
						;        0x000008B for 0.1 Hz
	
	;; SEC3 Constants definitions
;; These are the coefficients of the polynome transforming the triangle into sine wave.
;; The coefficients are used in instructions using S3.18 format.
Coeffx1	EQU	 0x06487E		; = pi/2
Coeffx3	EQU	 0x3D6F03		; = (5/2 - pi)
Coeffx5	EQU	 0x00487E		; = (pi/2 - 3/2)

	;; SEC4 Memory variables definition
ABS	Up		0x2
ABS	Value	0x3
ABS	Tmp		0x4
	
	;; SEC5 Initial values of memory variables
IUp		EQU	0x1000000
IValue	EQU	0x0
	
	;; SEC6 Initialization
CM	0x40000	INIT
SKIP	!Z	EndInit
C	IValue			; load value
SCA	0x0 Value
C	IUp			; load value
SCA	0x0 Up
C	IAMpl			; load value
SCA	0x0 Ampl
C	IDelta			; load value
SCA	0x0 Delta
C       0x1000000     ; 1->A
SCA     0x0     INIT        ; do not reinitialize next time
EndInit:

	;; SEC7 Module code
CM	0x40000 Value		; Value -> A
XCM	0x40000 Delta		; Value -> B, Delta -> A
AMB	Up					; Value +/- Delta -> A
AMC	0x3c0000 Up			; |'new'-Value| - 1 -> A
SKIP	N+Z Output
	;; we have  |overflow| in A
	;; we must 1) invert up
	;;         2) put |overflow|-2 in A
XCM		0x3c0000 Up		; save |overflow| in B, load -up
S1BC	0x380000 Up		; Save new up value, |overflow|-2 -> A 
Output:
DAC		0x100 	0x40000	; A+1
AMC		0x0 	Up      ; New A calculated
; transform now the triangle into a sine by applying the polynome
SXAAC	0.0	  	Value	; update value
						; A = x^2					B = x
AAC		0.0	  			; A = x^4
BAC		0.0	  			; A = x^5
CAD		Coeffx5	0.0		; A = 0.25 * x^5
SBBC	0.0	  	Tmp		; A = x^2					Tmp = 0.25 * x^5
BAC		0.0	  			; A = x^3
CAM		Coeffx3	Tmp		; A = -0.5 * x^3 + 0.25 * x^5
SCBA 	Coeffx1	Tmp		; A = 1.5 * x - 0.5 * x^3 + 0.25 * x^5
AMC		0.0   	Ampl	; rescale
SCA		1.0	  	OUT1	; write output
