;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Effect Name : Triangle to Sinus converter
;
; Description : Converts a triangle signal into a sine wave
;
; Nb In       : 1
; Nb Out      : 1
;
; Input       : triangle input
; Output      : sine wave output
;
; Cycles      : 11 + 6 cycles (-1=unknownw)
; Memory      : 2 words (without init)
; Delay Line  : no
;
; Author      : Lall
; Date        : 18/04/2003
; Version     : 1.0
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; The sinusoid is generated by generating a triangle and then
;;    applying a polynome of the following form to it:
;;    out = Coeff1 * in + Coeff3 * in^3 + Coeff5 * in^5
;;
;; The approximated sinus generates a contribution at 3*Fc and 5*Fc
;;    for 440Hz operating frequency and 48000 sampling frequency,
;;    the magnitude of the contributions are resp. -96dB and -85dB.
;;
;; Important remark: the amplitude parameter should be set to 0.5
;;    for full-scale operation when the output is directly
;;    connected to a DAC as the DAC is expecting 0.5 as full-scale.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


ABS	INIT	0x3FF
	
	;; SEC1 Control parameters
ABS	Ampl	0x0		; amplitude
	
	;; SEC2 Initial values of control parameters (S3.24 format)
IAmpl	EQU	0.5
	
	;; SEC3 Constants definitions
;; These are the coefficients of the polynome transforming the triangle into sine wave.
;; The coefficients are used in instructions using S3.18 format.
Coeffx1	EQU	 0x06487E		; = pi/2
Coeffx3	EQU	 0x3D6F03		; = (5/2 - pi)
Coeffx5	EQU	 0x00487E		; = (pi/2 - 3/2)

	;; SEC4 Memory variables definition
ABS	Tmp		0x1				; Uninitialized temporary memory
	
	;; SEC5 Initial values of memory variables
	
	;; SEC6 Initialization
CM	0x40000	INIT
SKIP	!Z	EndInit
C	IAmpl			; load value
SCA	0.0 Ampl
C       0x1000000     ; 1->A
SCA     0x0     INIT        ; do not reinitialize next time
EndInit:

	;; SEC7 Module code
CM	1.0		 	IN1		; Value -> A
; transform now the triangle into a sine by applying the polynome
SXAAC	0.0	  	Tmp		; dummy update Tmp
						; A = x^2					B = x
AAC		0.0	  			; A = x^4
BAC		0.0	  			; A = x^5
CAD		Coeffx5	0.0		; A = 0.25 * x^5
SBBC	0.0	  	Tmp		; A = x^2					Tmp = 0.25 * x^5
BAC		0.0	  			; A = x^3
CAM		Coeffx3	Tmp		; A = -0.5 * x^3 + 0.25 * x^5
SCBA 	Coeffx1	Tmp		; A = 1.5 * x - 0.5 * x^3 + 0.25 * x^5
AMC		0.0   	Ampl	; rescale
SCA		1.0	  	OUT1	; write output
