/////////////////////////////////////////////////////////////////////////////
// Name:        glmodule.cpp
// Project:     GUI for Axoris AL3101 application generator
// Author:      Gregor(y) Soyez
// Version:     1.1
// Created:     06/01/02
// Copyright:   (c) Axoris
/////////////////////////////////////////////////////////////////////////////

#include "glmodule.h"
#include "AL3101EditProperties.h"
#include "gldefs.h"

// ctor & dtor
CglModule::CglModule(wxWindow *i_parent, wxString i_Name, 
		     wxString i_Path, wxString i_Label, 
		     TTheme *i_theme,
		     int x_pos, int y_pos,
		     TModuleInfo *i_info){
#ifdef DEBUG
  fprintf(stderr, "DBG: glmodule: Entering constructor\n");
#endif

  Parent = i_parent;
  Name   = i_Name;
  Label  = i_Label;
  Path   = i_Path;

  NbProperties = 0;
  mpparam = NULL;
  
  x = x_pos;
  y = y_pos;

  strcpy(info.Name, i_info->Name);
  strcpy(info.Descr, i_info->Descr);
  strcpy(info.InDescr, i_info->InDescr);
  strcpy(info.OutDescr, i_info->OutDescr);
  info.NbIn = i_info->NbIn;
  info.NbOut = i_info->NbOut;
  info.Cycles = i_info->Cycles;
  info.InitCycles = i_info->InitCycles;
  info.MemUsage = i_info->MemUsage;
  info.UseDelay = i_info->UseDelay;

  // create temporary DC for bitmaps.
  theme = i_theme;
  UpdateSize();
  UpdateTextures();

  //  TexIO  = i_TexIO;

  State = MODULE_STATE_STATIC;
}

CglModule::~CglModule(){
#ifdef DEBUG
  fprintf(stderr, "DBG: glmodule: Entering destructor\n");
#endif

  if (NbProperties>0) delete[] Properties;
  DeleteParamList(&mpparam);
}

void CglModule::UpdateSize(){
  int NameWidth, NameHeight, oldnx;

  oldnx = nx;

  nx = (info.NbIn>info.NbOut) ? info.NbIn : info.NbOut;
  ny = 2;

  // Check if the space needed to write module name is big enough
  // If it is not the case, enlarge the module size
  wxClientDC dc(Parent);
  wxFont Font(FONT_SIZE, wxROMAN, wxNORMAL, wxNORMAL);
  if (Font!=wxNullFont){
    dc.SetFont(Font);
    dc.GetTextExtent(Label, &NameWidth, &NameHeight);
    NameWidth=(int) ((NameWidth+0.4*SIZE_OF_GRID)/SIZE_OF_GRID)+1;
    if (nx<NameWidth) nx=NameWidth;
  }

  if (oldnx!=nx) UpdateTextures();
}


void CglModule::UpdateTextures(){
  char tmp[255];

  sprintf(tmp, "../resources/AgALag/textures/%s", theme->TexMod);
  TexBk.LoadFile(tmp, wxBITMAP_TYPE_BMP);
  TexBk = TexBk.Rescale(SIZE_OF_GRID*nx, 2*SIZE_OF_GRID);

  sprintf(tmp, "../resources/AgALag/textures/%s", theme->TexPin);
  TexIO.LoadFile(tmp, wxBITMAP_TYPE_BMP);
  TexIO = TexIO.Rescale(SIZE_OF_GRID, SIZE_OF_GRID);

  if ((!TexBk.Ok()) || (!TexIO.Ok()))
    wxMessageBox("Incorrect textures", "Warning", wxOK);
}

// renderize
void CglModule::Render(wxDC *dc, int UseTextures){
  int i, i_offset;
  int s = SIZE_OF_GRID; //, ds=(int) (0.2*SIZE_OF_GRID), ls=(int) (0.6*SIZE_OF_GRID);

  dc->SetPen(wxPen(wxColour(theme->NameColor[0],
			    theme->NameColor[1],
			    theme->NameColor[2]),
		   1, wxSOLID));

  // draws main box
  if (UseTextures)
    dc->DrawBitmap(TexBk.ConvertToBitmap(),s*x,s*y);
  else {
    dc->SetPen(wxPen(wxColour(0,0,255), 1, wxSOLID));
    dc->SetBrush(wxBrush(wxColour(255,255, 255), wxSOLID));
    dc->DrawRectangle(s*x, s*y, s*nx, s*ny);
  }

  i_offset=0;
  // draws inputs and outputs
  if (UseTextures){
    for (i=i_offset;i<info.NbIn+i_offset;i++){
      dc->DrawBitmap((TexIO.Mirror(FALSE)).ConvertToBitmap(),s*(x+i),s*(y-1));
    }
    for (i=i_offset;i<info.NbOut+i_offset;i++){
      dc->DrawBitmap(TexIO.ConvertToBitmap(),s*(x+i),s*(y+ny));
    }
  } else {
    int ds=(int) (0.2*SIZE_OF_GRID), ls=(int) (0.6*SIZE_OF_GRID);
    dc->SetBrush(wxBrush(wxColour(128, 0, 128), wxSOLID));
    for (i=i_offset;i<info.NbIn+i_offset;i++){
      dc->DrawRectangle(s*(x+i)+ds, s*y, ls, -ls);
    }
    for (i=i_offset;i<info.NbOut+i_offset;i++){
      dc->DrawRectangle(s*(x+i)+ds, s*(y+ny), ls, ls);
    }
  }

  // write title
  dc->SetFont(wxFont(FONT_SIZE, wxROMAN, wxNORMAL, wxNORMAL));
  if (UseTextures)
    dc->SetTextForeground(wxColour(theme->NameColor[0],
				   theme->NameColor[1],
				   theme->NameColor[2]));
  else dc->SetTextForeground(wxColour(0,255,0));
  dc->DrawText(Label, s*x+5, s*y+3);

}

int CglModule::IsInside(int xp, int yp, int *io){
  int i_offset;

  if ((xp>=x) && (xp<(x+nx)) && (yp>=y) && (yp<(y+ny)))
    return MODULE_INSIDE;

  i_offset=0; //(nx-nout)/2;
  if ((yp==y+ny) && (xp>=x+i_offset) && (xp<(x+info.NbOut+i_offset))){
    (*io)=xp-x+1-i_offset;
    return MODULE_OUTPUT;
  }

  i_offset=0; //(nx-nin)/2;
  if ((yp==y-1) && (xp>=x+i_offset) && (xp<(x+info.NbIn+i_offset))){
    (*io)=xp-x+1-i_offset;
    return MODULE_INPUT;
  }

  return MODULE_OUTSIDE;
}


// methods for handling module movement
//   1. change position during movement
int CglModule::MousePositionChanged(int new_x, int new_y){
  if (State!=MODULE_STATE_MOVING){
    return MODULE_ERROR;
  }

  new_x-=x_offset;
  new_y-=y_offset;

  if ((new_x!=x) || (new_y!=y)){
    x=new_x;
    y=new_y;
    return MODULE_RENDER;
  } else {
    x=new_x;
    y=new_y;
    return MODULE_OK;
  }

  return MODULE_ERROR;
}

//   2. beginning of the movement
int CglModule::StartMove(int mx, int my){
#ifdef DEBUG
  fprintf(stderr, "DBG: glmodule: Entering StartMove\n");
#endif

  State = MODULE_STATE_MOVING;

  x_offset = mx-x;
  y_offset = my-y;

  return 0;
}

//   3. End of the movement
int CglModule::FinishMove(){
#ifdef DEBUG
  fprintf(stderr, "DBG: glmodule: Entering FinishMove\n");
#endif

  State = MODULE_STATE_STATIC;

  return 0;
}

// set index or the module and retreive it.
void CglModule::SetIndex(int i_Index){
#ifdef DEBUG
  fprintf(stderr, "DBG: glmodule: Entering SetIndex\n");
#endif

  Index=i_Index;
}

int CglModule::GetIndex(){
#ifdef DEBUG
  fprintf(stderr, "DBG: glmodule: Entering GetIndex\n");
#endif

  return Index;
}

// get module name
wxString CglModule::GetName(){
#ifdef DEBUG
  fprintf(stderr, "DBG: glmodule: Entering GetName\n");
#endif

  return Name;
}

// get module name
wxString CglModule::GetPath(){
#ifdef DEBUG
  fprintf(stderr, "DBG: glmodule: Entering GetPath\n");
#endif

  return Path;
}


// retreive property list for this module
int CglModule::RetreivePropertyList(char *dir){
  char FName[255];
  FILE *flux;

  sprintf(FName, "%s/%s%s.asm", dir, Path.c_str(), Name.c_str());
  flux=fopen(FName, "r");
  if (flux==NULL) return MODULE_ERROR;

  NbProperties=GetPropertyArray(flux, &Properties);

  fclose(flux);

  return MODULE_OK;
}


// change the properties of the module
int CglModule::UpdateProperties(){
  if (NbProperties>0){
    CProperties EditProps(Parent, -1, Properties, NbProperties);
    int i;
    if (EditProps.ShowModal()==wxID_OK){
      for (i=0;i<NbProperties;i++){
	EditProps.sProp[i].MakeUpper();
	strcpy(Properties[i].Value, EditProps.sProp[i].c_str());
	ExtractHexa(Properties[i].Value, &(Properties[i].value), 24);
      }
      return 1;
    }
  }
  return 0;
}


// save property list to the given script file
int CglModule::SaveProperties(FILE *flux){
  int i;

  for (i=0;i<NbProperties;i++){
    fprintf(flux, "%d %d %s\n", Index, i+1, Properties[i].Value);
  }

  return MODULE_OK;
}


// set a particular property
int CglModule::SetProperty(int i, char *val){
  if (i>NbProperties) return MODULE_ERROR;

  TParameterItem *item;
  int j;
  
  strcpy(Properties[i-1].Value, val);
  for (j=0;j<strlen(Properties[i-1].Value);j++) 
    Properties[i-1].Value[j] = toupper(Properties[i-1].Value[j]);
  ExtractHexa(Properties[i-1].Value, &(Properties[i-1].value), 24);

  // assign the corresponding parameter item
  item = mpparam;
  while ((item!=NULL) 
	 && (item->ParamAddress[0]!=Properties[i-1].address)){ 
    item = item->next; 
  };
  
  if (item!=NULL){
    item->ParamValue[0]=item->DisplayInitialValue=
      ConvertToFloat(Properties[i-1].value);
  }
  return MODULE_OK;
}


int CglModule::BuildDefaultParamList(bool Enable){
#ifdef DEBUG
  fprintf(stderr, "DBG: BuildParamList: Entering\n");
#endif
  int i;

  TParameterItem *item;
  mpparam=NULL;
  
  //wxLogMessage("Using default values");
  
  for (i=0;i<NbProperties;i++){
    item = new TParameterItem;
    item->next = mpparam;
    mpparam = item;
    strcpy(item->ParameterName, Properties[i].Name);
    strcpy(item->ParameterDescription, Properties[i].Descr);
    item->ParamAddress[0]=(int) Properties[i].address;
    AssignDefaultValues(item);
    item->ParamValue[0]=item->DisplayInitialValue=
      ConvertToFloat(Properties[i].value);
    item->InGroup = 0;
    item->Enabled = Enable;
  }

  return 0;
}

int CglModule::BuildParamList(){
#ifdef DEBUG
  fprintf(stderr, "DBG: BuildParamList: Entering\n");
#endif
  int i;

  TParameterItem *item;
  TParameterItem *item_prev;
  TProjectParameters *pl;
  TProjectParameters *pl_tmp;
  char pname[1024];

  // check if there is a MP project for this module
  CConfig *cfg;
  cfg = new CConfig;
  sprintf(pname, "%s/%s%s.prj", cfg->EffectsDir, GetPath().c_str(), GetName().c_str());
  delete cfg;
  if (LoadProject(pname, &pl)==0){
    // check that list is not empty
    if ((pl==NULL) || (pl->param==NULL)) return BuildDefaultParamList(false);

    // first param = first item of the first list !
    mpparam = pl->param;
    pl_tmp = pl;

    while (pl!=NULL){
      item = pl->param;
      if (item!=NULL){
	do{
	  item->Enabled = true;
	  // offset if already included in LoadProject
	  //item->ParamAddress[0]+=pl->AddressOffset;
	  item_prev = item;
	  item = item->next;
	} while (item!=NULL);
      }
      // link with the next param list
      if (pl->next!=NULL)
	item_prev->next = pl->next->param;
      pl=pl->next;
    };

    // release lists
    while (pl_tmp!=NULL){
      pl = pl_tmp->next;
      delete pl_tmp;
      pl_tmp = pl;
    };

  } else return BuildDefaultParamList(false);

  return 0;
}
