//--------------------------------------------------------------------//
//  AgALag : Axoris group AL3101 application generator                //
//                                                                    //
//                 _         _    _      _                            //
//                / \____   / \  | |    / \____                       //
//               / _ \  _\ / _ \ | |   / _ \  _\                      //
//              / /_\ \  __ /_\ \| |  / /_\ \  __                     //
//             / _____ \ \ \____ \ |_/______ \ \ \                    //
//            /_/     \_\___\   \_\____|    \_\___\                   //
//                                                                    //
//  File: structure.cpp                                               //
//  Description: manager for all modules inside the project           //
//  Version: 1.0                                                      //
//  Author: Gregor                                                    //
//  e-mail: Gregor.Soyez@swing.be                                     //
//                                                                    //
//  Copyright (C) 2002 Axoris                                         //
//--------------------------------------------------------------------//

#include <stdio.h>
#include <string.h>
#include "structure.h"
#include "AL3101ModuleHeader.h"
#include "ProjAL.h"
#include "modorder.h"
#include "module.h"
#include "memory.h"
#include "map.h"
#include "table.h"
#include "messages.h"
#include "iomapping.h"

/*
** Creates the header of the asm file
*/
int CreateHeader(FILE *flux, TModuleInfo *pInfo){
  fseek(flux, 0L, SEEK_SET);
  fprintf(flux, ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n");
  fprintf(flux, "; Effect Name : %s\n", pInfo->Name);
  fprintf(flux, ";\n");
  fprintf(flux, "; Description : %s\n", pInfo->Descr);
  fprintf(flux, ";\n");
  fprintf(flux, "; Nb In       : %d\n", pInfo->NbIn);
  fprintf(flux, "; Nb Out      : %d\n", pInfo->NbOut);
  fprintf(flux, ";\n");
  fprintf(flux, "; Input       : %s\n", pInfo->InDescr);
  fprintf(flux, "; Output      : %s\n", pInfo->OutDescr);
  fprintf(flux, ";\n");
  fprintf(flux, "; Cycles      : %d + %d cycles\n", pInfo->Cycles, pInfo->InitCycles);
  fprintf(flux, "; Memory      : %d words (without init)\n", pInfo->MemUsage);
  fprintf(flux, "; Delay Line  : %s\n", ((pInfo->UseDelay==1) ? "yes" : "no"));
  fprintf(flux, ";\n");
  fprintf(flux, "; Author      : AgALag (Axoris application generator)\n");
  fprintf(flux, "; Date        : \n");
  fprintf(flux, "; Version     : \n");
  fprintf(flux, ";\n");
  fprintf(flux, ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n");
  fprintf(flux, "\n");
  fprintf(flux, ";; INIT memory address\n");
  fprintf(flux, "ABS INIT 0x3FF\n");
  fprintf(flux, "\n");

  return 0;
}

/*
** Creates the header of the given section
*/
int CreateSectionHeader(FILE *flux, int sec){
  switch (sec){
  case 1:
    fprintf(flux, "\t;; SEC1 Control parameters\n");
    break;
  case 2:
    fprintf(flux, "\t;; SEC2 Initial values of control parameters (S3.24 format)\n");
    break;
  case 3:
    fprintf(flux, "\t;; SEC3 Constants definitions\n");
    break;
  case 4:
    fprintf(flux, "\t;; SEC4 Memory variables definition\n");
    break;
  case 5:
    fprintf(flux, "\t;; SEC5 Initial values of memory variables\n");
    break;
  case 6:
    fprintf(flux, "\t;; SEC6 Initialization\n");
    fprintf(flux, "CM      0x40000 INIT\n");
    fprintf(flux, "SKIP    !Z      EndInit\n");
    break;
  case 7:
    fprintf(flux, "EndInit:\n");
    fprintf(flux, "\n");
    fprintf(flux, "\t;; SEC7 Module code\n");
    break;
  };

  return 0;
}

/*
** This function takes a .axp file and creates
** the .asm file with the application code.
** it returns 0 if successful and 1 otherwise.
*/
int MakeApplication(char *ProjectFile){
  char AsmFile[255];
  FILE *flux;
  CCodeModule **Mods;
  int NbMod, i, j, k, l, tmp, len;
  int UseDelayLine;
  int InputIndex, OutputIndex;
  TModuleInfo ProjectInfo;

  // STEP 0: check the profect file name andf create the appl file name
  len=strlen(ProjectFile);
  if ((ProjectFile[len-4]!='.') || (ProjectFile[len-3]!='a') 
      || (ProjectFile[len-2]!='x') || (ProjectFile[len-1]!='p')){
    ERROR1("%s is not a .axp file ! Giving up.\n", ProjectFile);
    return 1;
  }
  strcpy(AsmFile, ProjectFile);
  AsmFile[len-2]='s';
  AsmFile[len-1]='m';

  // STEP 1: read the project file and arrange the modules into a structure
  
  // retreive the number of modules
  if (CreateStructure(ProjectFile, NULL, &NbMod, &ProjectInfo) != MODORDER_OK)
    return 1;

  Mods=new CCodeModule *[NbMod];
  for (i=0;i<NbMod;i++)
    Mods[i]=new CCodeModule();

  if (CreateStructure(ProjectFile, (CModule**) Mods, &NbMod, &ProjectInfo) != MODORDER_OK){
    for (j=0;j<NbMod;j++)
      delete Mods[j];
    delete[] Mods;
    return 1;
  }

  // STEP 2: for each module (except IN & OUT) locate sections 
  OUTPUT("Scanning sections in source files.\n");
  InputIndex=OutputIndex=-1;
  for (i=0;i<NbMod;i++){
    if (strcmp(Mods[i]->Name, OUTPUT_MODULE_NAME)==0){
      OutputIndex=i;
    } else if (strcmp(Mods[i]->Name, INPUT_MODULE_NAME)==0){
      InputIndex=i;
    } else {
      if (Mods[i]->CheckSections()!=MODULE_OK){
	for (j=0;j<NbMod;j++)
	  delete Mods[j];
	delete[] Mods;
	return 1;
      }
    }
  }
  if ((InputIndex==-1) || (OutputIndex==-1)){
    ERROR("One of the IO modules is missing ! Giving up.\n");
    for (i=0;i<NbMod;i++)
      delete Mods[i];
    delete[] Mods;
    return 1;
  }


  // STEP 3: deal with memory allocation
  OUTPUT("Determination of memory offsets\n");
  MemAllocInit();
  UseDelayLine=-1;
  // scanning if one need a delayline
  // that module will be placed at the last memory segment to allow for maximal delay
  for (j=0;j<NbMod;j++){
    if ((j!=InputIndex) && (j!=OutputIndex)){
      if (Mods[j]->GetDelayLine()==1){
	if (UseDelayLine!=-1){
	  // only support one delay line for the moment
	  ERROR("More than one module needs a delay line! Giving up.\n");
	  for (i=0;i<NbMod;i++)
	    delete Mods[i];
	  delete[] Mods;
	  return 1;
	} else {
	  UseDelayLine=j;
	}
      }
    }
  }
  
  // browsing the other ones
  for (j=0;j<NbMod;j++){
    if ((j!=InputIndex) && (j!=OutputIndex) && (j!=UseDelayLine)){
      len=SegAlloc(Mods[j]->GetMemoryNeeds());
      if (len==-1){
	for (i=0;i<NbMod;i++)
	  delete Mods[i];
	delete[] Mods;
	return 1;
      }
      Mods[j]->SetMemOffset(len);
    }
    if ((j==InputIndex) || (j==OutputIndex)){
      Mods[j]->info.Cycles=Mods[j]->info.InitCycles=Mods[j]->info.MemUsage=Mods[j]->info.UseDelay=0;
    }
  }

  // calculating total number of cycles, memory needs 
  // Note that we already know if we need a delay line
  ProjectInfo.Cycles=0;
  ProjectInfo.InitCycles=4;
  for (i=0;i<NbMod;i++){
    ProjectInfo.Cycles+=Mods[i]->info.Cycles;
    if (Mods[i]->info.InitCycles>4) ProjectInfo.InitCycles+=Mods[i]->info.InitCycles-4;
  }
  ProjectInfo.MemUsage=SegAlloc(0);

  // STEP 4: For each module creates the temporary IO 
  OUTPUT("Creating temporary IO\n");

  // we can use the free adresses
  InitIOMapping(ProjectInfo.MemUsage);

  for (j=0;j<NbMod;j++){
    if ((j!=OutputIndex)){
      VERBOSE1("  module %s : ", Mods[j]->Name);
      switch (Mods[j]->CreateTemporaryIO()){
      case MODULE_OK:
	VERBOSE("Ok\n");
	break;
      case MODULE_NO_TMP_LEFT:
	ERROR1("%s: Not enough temporary IO available\n", Mods[j]->Name);
	for (i=0;i<NbMod;i++)
	  delete Mods[i];
	delete[] Mods;
	return 1;
	break;	
      case MODULE_ADD_ERROR:
	ERROR1("%s: Error when appending a replacement to list\n", Mods[j]->Name);
	for (i=0;i<NbMod;i++)
	  delete Mods[i];
	delete[] Mods;
	return 1;
	break;	
      case MODULE_WRONG_STRUCTURE:
	ERROR1("%s: Invalid IO Structure\n", Mods[j]->Name);
	for (i=0;i<NbMod;i++)
	  delete Mods[i];
	delete[] Mods;
	return 1;
	break;	
      };
      // check if we can free some IO's
      for (k=0;k<GetNbMapped();k++){
	tmp=0;
	for (i=j+1;i<NbMod;i++)
	  for (l=0;l<2*NB_CHANNELS;l++)
	    if (Mods[i]->CatchedTmpIO[l]==k) tmp=1;
	if (!tmp) 
	  if (FreeIO(k)==IO_MAPPING_OK)
	    VERBOSE1("  Freeing tempoprary io %d\n", k);
      }
    }
  }

  // allocate memory for temporary IO
  SegAlloc(GetNbMapped());
  ProjectInfo.MemUsage+=GetNbMapped();

  // STEP 5 deal with memory allocation for delay line
  // allocating memory for module with delay line
  // telling user the maximum size for the delay line (if delay line!)
  if (UseDelayLine!=-1){
    len=SegAlloc(Mods[UseDelayLine]->GetMemoryNeeds());
    Mods[UseDelayLine]->SetMemOffset(len);
    // len max for delay = 1024-len-2
    // the -2 takes into account the INIT and the address pointer
    len=1024-len-2;
    ProjectInfo.UseDelay = 1;
    OUTPUT2("The available memory for delay line is %d samples (%4.1f ms)\n", len, (len/48.0));
  } else {
    ProjectInfo.UseDelay = 0;
  }

  // STEP 6: create the source code by adding modules in the correct order
  //         section by section;
  OUTPUT("Creating assembler code\n");
  flux=fopen(AsmFile, "w+");

  // count the number of  "SCA ... INIT" from init sections
  //for (j=0;j<NbMod;j++)
  //  if ((j!=InputIndex) && (j!=OutputIndex))
  //    ProjectInfo.InitCycles-=Mods[j]->InitRemoved();
  //ProjectInfo.InitCycles++;

  // check that we do not have more than 256 init cycles and total number of cycles <= 1024
  // note that we cannot do that before due to the fact that lones like "SCA ... INIT" are removed only at this stage
  if ((ProjectInfo.InitCycles>255) && (!Options.GenerateMem)){
    ERROR("The total number of init cyles exceeds maximal value (255)\nYou can solve this problem by generating a memory map.\n");
    for (i=0;i<NbMod;i++)
      delete Mods[i];
    delete[] Mods;
    return 1;
  }
  if (((ProjectInfo.InitCycles+ProjectInfo.Cycles>1024) && (!Options.GenerateMem))
      || (ProjectInfo.Cycles>1024)){
    ERROR("The total number of cyles exceeds maximal value (1024)\nYou can probably solve this problem by generating a memory map.\n");
    for (i=0;i<NbMod;i++)
      delete Mods[i];
    delete[] Mods;
    return 1;
  }

  // header
  CreateHeader(flux, &ProjectInfo);

  // sections one by one
  for (i=1;i<=NB_SECTIONS;i++){
    if ((Options.GenerateMem) && ((i==2) || (i==5) || (i==6))){
      VERBOSE1("Skipping  section %d\n", i);
    } else {
      VERBOSE1("Appending section %d\n", i);
      CreateSectionHeader(flux, i);
      for (j=0;j<NbMod;j++){
	if ((j!=InputIndex) && (j!=OutputIndex)){
	  VERY_VERBOSE1("  module %s : ", Mods[j]->Name);
	  if (Mods[j]->AppendSectionToStream(i, flux)!=MODULE_OK){
	    ERROR1("%s: Syntax error\n", Mods[j]->Name);
	    fclose(flux);
	    // remove asm file
	    remove(AsmFile);
	    // release memory
	    for (i=0;i<NbMod;i++)
	      delete Mods[i];
	    delete[] Mods;
	    return 1;
	  } else {
	    VERY_VERBOSE("Ok\n");
	  }
	} // ends 'if'
      } // ends modules loop
      if (i==4) WriteMappingDefs(flux);
      if (i==6){
	fprintf(flux,"C 0x1000000\t; Set A to 1\n");
	fprintf(flux,"SCA 0x0 INIT\t; do not reinitialise next time\n\n");
      }
      if (i<6) fprintf(flux, "\n");
    } // ends section stuff
  }

  fclose(flux);

  // STEP 7: generate memory map if requested
  if (Options.GenerateMem){
    OUTPUT("Generating Memory Map\n");
    GenerateMemoryMap(ProjectFile, Mods, NbMod);
  }

  // STEP 8: generate offset table
  if (Options.GenerateTable){
    OUTPUT("Generating Offset table\n");
    GenerateOffsetTable(ProjectFile, Mods, NbMod);
  }

  // STEP 9: release memory

  FinishIOMapping();
  for (i=0;i<NbMod;i++)
    delete Mods[i];
  delete[] Mods;

  OUTPUT("Success.\n");

  return 0;
}
