/* Assembler for AL3101 */

#include <stdio.h>
#include <string.h>

#include "./../MessageMacros.h"
#include "AL3101Table.h"
#include "AL3101Struct.h"
#include "AL3101DbgFiles.h"
#include "AL3101AsmDisasmUtils.h"

#include "AL3101PreAsm.h"
#include "AL3101Asm.h"

#define NB_MAX_COMMAND    1024  // maximal number of commands allowed by the chip.
#define MAX_TOKEN_SIZE      80  // Maximum allowed size of a token
#define MAX_CMDLINE_SIZE   255  // Maximum allowed size of a command line
#define NB_MAX_TOKENS        3  // mnemonic + 2 param

#define ASM_PARSE_ERROR     -1  // parsing error
                                // only used internal so ... 
                                //   doesn't need to appear in .h

// Mnemonic Table defined in "AL3101Table.cpp"
extern TMnemonicTable MnemonicTable[];

typedef struct{
  int LineIndex;     // reference of the line in the .asm file (only usefull for error reporting)
  char Command[MAX_TOKEN_SIZE];
  char *Params[NB_MAX_TOKENS-1];
  int NbParams;
} TPreAssembledCommand;


/*
** Check the given filename :
**  1. add .asm extension if not present
**  2. check if file exist
** This function returns 0 on succeed, 1 otherwise
*/
int CheckAsmFilename(char* fn){
  int len = strlen(fn);
  
  // check if '.asm' is present at end of fn. Add '.asm' if it is not the case
  if ((len<5) || (fn[len-1]!='m') || (fn[len-2]!='s') 
      || (fn[len-3]!='a') || (fn[len-4]!='.')){
    strcat(fn, ".asm");
    // don't forget to update the length of the string
    len+=4;
  }

  // check if the file exist
  FILE *flux=fopen(fn, "r");
  if (flux==NULL)
    return 1;
  fclose(flux);

  return 0;
}

// Parse command line into mnemonic and parameters
// returns the number of tokens found or ASM_PARSE_ERROR if error
int Parse(char *CmdLine, char *Token[]){
  int start, end;
  int n, i;

  // retreive the length of the command line
  n=strlen(CmdLine);

  // if CommandLine is an emtpy string, return 0
  if (n==0)
    return 0;

  // if there is(are) EOL caracter(s), replace it(them) by null caracters
  // We are taking into account two different types of EOL caracters :
  //   0x0A and 0x0D.
  // For windows, we should have both while under Linux only one of them 
  // should be present
  while ((CmdLine[n-1]==0x0A) || (CmdLine[n-1]==0x0D)){
    CmdLine[n-1]='\0';
    n--;
    // if CommandLine is an emtpy string, return 0
    if (n==0)
      return 0;
  }
  
  start=0;
  i=0;

  do{   // we loop until we arrive at EOL or when we find a comment sign
    // find the beginning of the next token
    while ((CmdLine[start]=='\t') || (CmdLine[start]==' ')) start++;

    // if EOL or comment, return the number of tokens found
    if ((start==n) || (CmdLine[start]==';')) return i;

    // if there is no more token allowed, we face a typical parse error
    if (i==NB_MAX_TOKENS) return ASM_PARSE_ERROR;

    // scan the beginning token ...
    end=start+1;
    while ((CmdLine[end]!='\t') && (CmdLine[end]!=' ') 
	   && (CmdLine[end]!=';') && (end<n)) end++;

    // ... and store it.
    strncpy(Token[i], CmdLine+start, end-start);
    Token[i][end-start]='\0';
    i++;

    // prepare for a next search.
    start=end; //+1;
  } while ((start<=n) && (i<=NB_MAX_TOKENS));

  return i; // returns the number of tokens found
}

/* Get the opcode associated with mnemonic mn.
** return 0 if succeeded, 1 otherwise
** if the function succeed, the opcode is 
** returned through 'op'
** The index in MnemonicTable is returned through 'index'
*/
int GetOPCode(char *mn, INT32 *op, int *index){
  int i=0;

  // browse the MnemonicTable commands
  do{
    if (strcmp(MnemonicTable[i].Mnemonic, mn)==0){
      (*index) = i;
      (*op) = MnemonicTable[i].OpCode;
      return 0;
    }      
    i++;
  } while (i<NB_OPCODES);

  return 1;
} 

/* If you try to write to addresses 0x410 to 0x417, 
** redirect it to 0x430 to 0x437
** This works for opcodes from to 0x100000 (SCA) to 0x110110 (S1BC).
** The "offset test" for SINDF and SINDI will de performed at runtime.
** This trick is used olny in the simulator version of the assembler,
** it is not part of the Alesis assembler syntax itself.
*/
int OutOffset(INT32 OpCode, INT32 *MemAddr){

  if (((OpCode<=54) && (OpCode>=32))
      && (((*MemAddr)<=1047) && ((*MemAddr)>=1040))){
    (*MemAddr)+=32;
  }

  return 0;
} 

/*
** Assemble the .asm file 'In_fn' and fill the structure AL3101_disasm
** Returns : NbLines             if success
**           ASM_FILE_NOT_EXISTS if In_fn doesn't exists
**           ASM_READ_ERROR      if an error occurs while browsing In_fn
**           ASM_ERROR           if errors were found in the source code
*/
int Assemble(char *In_fn, AL3101_Disasm_struct *AL3101_disasm){
  FILE *flux_in;                 // inout stream
  long int FileSize;             // Size of the input file
  char Input_fn[255];            // buffer to copy In_fn

  int ErrorNumber=0;             // Number of detected errors
  int LineIndex=0;               // index of the command line currently read
  int CmdIndex=0;                // index of the effective command 
  int NbCmd=0;                   // number of commands
  char CmdLine[255];             // command line in the input file
  char* Tokens[NB_MAX_TOKENS];   // Mnemonic+Params in the command line
  int nb_tok;                    // number of parameters + 1
  INT32 opcode;                  // opcode associated with Mnemonic
  int opcode_index;              // index of the opcode in the Mnemonictable
  int Code;                      // code for assembled parameters and instruction
  TPreAssembledCommand CCode[NB_MAX_COMMAND]; // store preassembled code
  

  // Copy the filename to a local variable.
  // This is useful when .asm has to be added. In such a case we have no
  // interference with the calling application.
  strcpy(Input_fn, In_fn);

  // check if it is a valid filename
  if (CheckAsmFilename(Input_fn)!=0){
    ERROR_1("File '%s' doesn't exist.\n", Input_fn);
    return ASM_FILE_NOT_EXISTS;
  }

  // open source file
  flux_in  = fopen(Input_fn,  "r");

  // allocate memory for future tokens
  sprintf(CmdLine, "");
  for (nb_tok=0;nb_tok<NB_MAX_TOKENS;nb_tok++) Tokens[nb_tok]=new char[MAX_TOKEN_SIZE];
    
  // calculating size of the input file
  fseek(flux_in, 0L, SEEK_END);
  FileSize=ftell(flux_in);
  fseek(flux_in, 0L, SEEK_SET);

  // Init pre-assembler
  InitPreAsm();

  /* 
  ** We perform assembling in two steps :
  ** In the first one, we read all lines from the file, for each of these, we tokenize it.
  ** We decide between 3 cases:
  **  1. comment or empty line (Nb_tokkens == 0)
  **  2. Preasm Command: this can be EQU, ABS or a label definition. We thus call the 
  **     preassembler during this first stage.
  **  3. Everything else should be a valid command and, unless the number of tokens is
  **     wrong (>3), we store command in the CCode array !
  ** The second step consist in assembling the CCode informationg through the "pure" assembler.
   */

  // STEP 1.

  // browse command lines until EOF reached or an error occured
  while ((NbCmd<1024) && (fgets(CmdLine, MAX_CMDLINE_SIZE, flux_in)!=NULL)){

    LineIndex++;

    // convert all upper case ...
    for (nb_tok=0;nb_tok<(int) strlen(CmdLine);nb_tok++){
      if ((CmdLine[nb_tok]>='a') && (CmdLine[nb_tok]<='z')){
	CmdLine[nb_tok]-=('a'-'A');
      }
    }

    // parse command line into tokens
    if ((nb_tok=Parse(CmdLine, Tokens))==ASM_PARSE_ERROR){
      ErrorNumber++;
      ERROR_1("%d: Parse error : too much paramerters\n", LineIndex);
      continue;
    }
    if (nb_tok==0) continue;

    // Pre-assemble string.
    switch(PreAssemble(nb_tok, Tokens, NbCmd)){
    case PREASM_NOTHING: // not a comment nor a preasm command -> store for step 2
      CCode[NbCmd].LineIndex=LineIndex;
      strcpy(CCode[NbCmd].Command, Tokens[0]);
      CCode[NbCmd].NbParams=nb_tok-1;
      CCode[NbCmd].Params[0]=new char[MAX_TOKEN_SIZE];
      CCode[NbCmd].Params[1]=new char[MAX_TOKEN_SIZE];
      for (nb_tok=0;nb_tok<CCode[NbCmd].NbParams;nb_tok++)
	strcpy(CCode[NbCmd].Params[nb_tok], Tokens[nb_tok+1]);
      NbCmd++;
      break;
    case PREASM_OK:      // preasm command treated by PreAssemble. Skip to next line.
      break;
    case PREASM_MAX_CST: // Preasm command ... preassembler capacity exceeded
      ErrorNumber++;
      ERROR_1("%d: Preassembler reached maximal capacity\n", LineIndex);
      break;
    case PREASM_SYNTAX:  // preassembler command with syntax error
      ErrorNumber++;
      ERROR_1("%d: Wrong preassembler syntax\n", LineIndex);
      break;
    case PREASM_FORMAT:  // preassembler command with format error
      ErrorNumber++;
      ERROR_1("%d: Wrong preassembler format\n", LineIndex);
      break;
    case PREASM_EXISTS:  // already defined preassembler constant
      ErrorNumber++;
      ERROR_2("%d: %s already defined\n", LineIndex, Tokens[0]);
      break;
    };

    sprintf(CmdLine, "");
  };

  // release memory for tokens
  for (nb_tok=0;nb_tok<NB_MAX_TOKENS;nb_tok++) delete[] Tokens[nb_tok];

  // if we have not read 'Filesize bytes, an error occured
  if (ftell(flux_in)<FileSize){
    fclose(flux_in);
    ERROR("Error before EOF ... Giving up.\n\n");
    return ASM_READ_ERROR;
  }

  // close file
  fclose(flux_in);

  // STEP 2

  for (CmdIndex=0;CmdIndex<NbCmd;CmdIndex++){

    // retreive opcode
    if (GetOPCode(CCode[CmdIndex].Command, &opcode, &opcode_index)!=0){
      ErrorNumber++;
      ERROR_2("%d: Unrecognized mnemonic '%s'\n", 
	      CCode[CmdIndex].LineIndex, CCode[CmdIndex].Command);
      continue;
    }

    // call assembling function associated with opcode
    // this fills the parameter fields in the AL3101_disasm structure passed
    // as argument
    switch ((Code=MnemonicTable[opcode_index].f_asm(CCode[CmdIndex].NbParams, &(CCode[CmdIndex].Params[0]), &(AL3101_disasm[CmdIndex])))){
    case -1:  // wrong number of parameter
      ERROR_3("%d: '%s' doesn't take %d parameters\n", 
	      CCode[CmdIndex].LineIndex, CCode[CmdIndex].Command, CCode[CmdIndex].NbParams);
      ErrorNumber++;
      break;
    case 1:   // error in parameter 1
    case 2:   // error in parameter 2
      ERROR_2("%d: Error in parameter %d\n", CCode[CmdIndex].LineIndex, Code);
      ErrorNumber++;
      break;
    default:   // call successful
      // Take into account Offset for OUT1->OUT8 addresses.
      SimulatorOutputOffset(opcode, &(AL3101_disasm[CmdIndex].m));
      // Fill missing fields
      AL3101_disasm[CmdIndex].op=opcode;
      AL3101_disasm[CmdIndex].s_index=opcode_index;
      AL3101_disasm[CmdIndex].ln=CCode[CmdIndex].LineIndex;
      // complete the binary code with the opcode
      AL3101_disasm[CmdIndex].wObj+=(opcode << 2);
      // storing initial command string
      sprintf(AL3101_disasm[CmdIndex].CmdStr, "%s\t", CCode[CmdIndex].Command);
      for (nb_tok=0;nb_tok<CCode[CmdIndex].NbParams;nb_tok++){
	strcat(AL3101_disasm[CmdIndex].CmdStr, CCode[CmdIndex].Params[nb_tok]);
	strcat(AL3101_disasm[CmdIndex].CmdStr, " ");
      }
      break;
    }

    // Update PreAsm Program Counter
    PreAsmStepPC();
  }

  // since memory desallocation must be done externally,
  // this line will be uncommented once the core of the simulator 
  // handle memory maps.
  //AL3101_disasm->NbMapItems = GenerateMap(&(AL3101_disasm->Map));

  // free memory allocate for parameters inside the CCode array
  for (CmdIndex=0;CmdIndex<NbCmd;CmdIndex++){
    delete[] CCode[CmdIndex].Params[0];
    delete[] CCode[CmdIndex].Params[1];
  }

  // Close PreAsm !
  PreAsmClose();

  // show number of errors
  if (ErrorNumber==0){
    ERROR("Assembling successfully completed\n\n");
    return NbCmd;
  } else {
    ERROR_1("Assembling found %d Errors\n\n", ErrorNumber);    
    return ASM_ERROR;
  }

  return NbCmd;
}

