#include <stdio.h>
#include <string.h>
#include "AL3101Struct.h"
#include "AL3101AsmFunc.h"
#include "AL3101PreAsm.h"
#include "AL3101AsmDisasmUtils.h"
#include "AL3101Format.h"

// maximal values for different kinds of constants
#define MAX_C18 0x3FFFFF
#define MAX_C24 0xFFFFFFF
#define MAX_C10 0x3FF
#define MAX_C8  0xFFF
#define MAX_D18 0x3FFFFF
#define MAX_D8  0xFFF
#define MAX_M   0x42F
#define MAX_P   0xF

// conditions
char* AsmConditionTable[] = {
  "!Z",
  "Z",
  "!V",
  "V",
  "N",
  "!N",
  "VN",
  "V!N",
  "!VN",
  "!V!N",
  "!V!N!Z",
  "!N!Z",
  "N+Z",
  "!V(N+Z)",
  "",
  "VZ"
};

// ASSEMBLING FUNCTIONS
// all the assembling functions have the same structure :
//   they take the number of parameters, and a pointer to 
//   the parameters (as strings) in entry. The returned value
//   is  0 if success,
//      -1 if the number of parameters is wrong
//       n if an error has been detected for the nth parameter
 
// normal mnemonics (no extrabit, no condition)
int asm_c18_x_m(int n, char **p, AL3101_Disasm_struct *Struct){
  INT32 c18, m;

  if (n!=2) return -1;
  if ((ExtractHexa(p[0], &c18, 18)!=0) || (c18>MAX_C18)) return 1;
  if ((ExtractHexa(p[1], &m  , -1)!=0) || (m  >MAX_M  )) return 2;

  Struct->c = (c18<<6);
  Struct->m = m;
  ExtendTo32Bits(&(Struct->c));

  // Binary code
  Struct->lObj = (c18<<12) + m;
  Struct->wObj = (c18>>20);
  return 0;
}

int asm_x_m(int n, char **p, AL3101_Disasm_struct *Struct){
  INT32 m;

  if (n!=1) return -1;
  if ((ExtractHexa(p[0], &m  , -1)!=0) || (m  >MAX_M  )) return 1;
  
  Struct->m = m;

  // Binary code
  Struct->lObj = m;
  Struct->wObj = 0;

  return 0;
}

int asm_c24_x(int n, char **p, AL3101_Disasm_struct *Struct){
  INT32 c24, r;

  if (n!=1) return -1;
  if (((r=ExtractHexa(p[0], &c24, 24))!=0) || (c24>MAX_C24)) return 1;

  Struct->c = c24;
  ExtendTo32Bits(&(Struct->c));

  // Binary code
  Struct->lObj = (c24<<6);
  Struct->wObj = (c24>>26);
  return 0;
}

int asm_c18_x(int n, char **p, AL3101_Disasm_struct *Struct){
  INT32 c18;

  if (n!=1) return -1;
  if ((ExtractHexa(p[0], &c18, 18)!=0) || (c18>MAX_C18)) return 1;

  Struct->c = (c18<<6);
  ExtendTo32Bits(&(Struct->c));

  // Binary code
  Struct->lObj = (c18<<12);
  Struct->wObj = (c18>>20);
  return 0;
}

int asm_c18_d8(int n, char **p, AL3101_Disasm_struct *Struct){
  INT32 c18, d8;

  if (n!=2) return -1;
  if ((ExtractHexa(p[0], &c18, 18)!=0) || (c18>MAX_C18)) return 1;
  if ((ExtractHexa(p[1], &d8 ,  8)!=0) || (d8 >MAX_D8 )) return 2;

  Struct->c = (c18<<6);
  Struct->d = (d8<<16);
  ExtendTo32Bits(&(Struct->c));
  ExtendTo32Bits(&(Struct->d));

  // Binary code
  Struct->lObj = (c18<<12) + d8;
  Struct->wObj = (c18>>20);
  return 0;
}

int asm_c8_d18(int n, char **p, AL3101_Disasm_struct *Struct){
  INT32 c8, d18;

  if (n!=2) return -1;
  if ((ExtractHexa(p[0], &c8 ,  8)!=0) || (c8 >MAX_C8 )) return 1;
  if ((ExtractHexa(p[1], &d18, 18)!=0) || (d18>MAX_D18)) return 2;

  Struct->c = (c8<<16);
  Struct->d = (d18<<6);
  ExtendTo32Bits(&(Struct->c));
  ExtendTo32Bits(&(Struct->d));

  // Binary code
  Struct->lObj = (c8<<22) + d18;
  Struct->wObj = (c8>>10);
  return 0;
}

int asm_x(int n, char **p, AL3101_Disasm_struct *Struct){
  if (n!=0) return -1;

  // Binary code
  Struct->lObj = 0;
  Struct->wObj = 0;
  return 0;
}

// mnemonics with condition
int asm_c24_x_p(int n, char **p, AL3101_Disasm_struct *Struct){
  INT32 c24, cdt, i;

  if ((n==0) || (n>2)) return -1;
  if (n==2){
    i=0;
    while ((i<0x10) && (strcmp(AsmConditionTable[i], p[0])!=0)) i++;
    if (i==0x10) return 1; else{
      cdt=i;
      i=1;
    }
  } else {
    i=0;
    cdt = 0xE;
  }
  if ((ExtractHexa(p[i], &c24, 24)!=0) || (c24>MAX_C24)) return (i+1);

  Struct->m = cdt;
  Struct->c = c24;
  ExtendTo32Bits(&(Struct->c));

  // Binary code
  Struct->lObj = (c24<<6) + cdt;
  Struct->wObj = (c24>>26);
  return 0;
}

int asm_c8_x_p(int n, char **p, AL3101_Disasm_struct *Struct){
  INT32 c8, cdt, i;

  if ((n==0) || (n>2)) return -1;
  if (n==2){
    i=0;
    while ((i<0x10) && (strcmp(AsmConditionTable[i], p[0])!=0)) i++;
    if (i==0x10) return 1; else{
      cdt=i;
      i=1;
    }
  } else {
    cdt=0xE;
    i=0;
  }
  // Note: here is the only command where eventual call to preasm ask for a label
  if ((ExtractHexa(p[i], &c8 , -2)!=0) || (c8 >MAX_C8 )) return (i+1);

  Struct->m = cdt;
  Struct->c = c8;

  // Binary code
  Struct->lObj = (c8<<26) + cdt;
  Struct->wObj = (c8>>6);
  return 0;
}

// mnemonics with extra bit
int asm_c10_x_0_m(int n, char **p, AL3101_Disasm_struct *Struct){
  INT32 c10, m;

  if (n!=2) return -1;
  if ((ExtractHexa(p[0], &c10,  0)!=0) || (c10>MAX_C10)) return 1;
  if ((ExtractHexa(p[1], &m  , -1)!=0) || (m  >MAX_M  )) return 2;

  Struct->m = m;
  Struct->c = c10;

  // Binary code
  Struct->lObj = (c10<<24) + m;
  Struct->wObj = (c10>>8);
  return 0;
}

int asm_c10_x_1_m(int n, char **p, AL3101_Disasm_struct *Struct){
  INT32 c10, m;

  if (n!=2) return -1;
  if ((ExtractHexa(p[0], &c10,  0)!=0) || (c10>MAX_C10)) return 1;
  if ((ExtractHexa(p[1], &m  , -1)!=0) || (m  >MAX_M  )) return 2;

  Struct->m = m;
  Struct->c = c10;

  // Binary code
  Struct->lObj = (c10<<24) + m + (1<<11);
  Struct->wObj = (c10>>8);
  return 0;
}

int asm_x_0_x(int n, char **p, AL3101_Disasm_struct *Struct){
  if (n!=0) return -1;

  // Binary code
  Struct->lObj = 0;
  Struct->wObj = 0;
  return 0;
}

int asm_c24_1_x(int n, char **p, AL3101_Disasm_struct *Struct){
  INT32 c24;

  if (n!=1) return -1;
  if ((ExtractHexa(p[0], &c24, 24)!=0) || (c24>MAX_C24)) return 1;

  Struct->c = c24;
  ExtendTo32Bits(&(Struct->c));

  // Binary code
  Struct->lObj = (c24<<6) + (1<<5);
  Struct->wObj = (c24>>26);
  return 0;
}
