/* Assembler for AL3101 */

#include <stdio.h>
#include <string.h>

#include "./../MessageMacros.h"
#include "AL3101Table.h"
#include "AL3101Disasm.h"
#include "AL3101AsmDisasmUtils.h"

#define MAX_CMDLINE_SIZE   255

extern TMnemonicTable MnemonicTable[];

char Comments[64][50]={
  "     C *[M]",            // CM
  "     C *[M], [M] -> B",  // LCM
  " A + C *[M]",            // CMA
  " A + C *[M], [M] -> B",  // LCMA
  " B + C *[M]",            // CMB
  "     C *[M],   A -> B",  // XCM
  " B + A *[M]",            // AMB
  " A + C *[M],   A -> B",  // XCMA
  " C + A *[M]",             // AMC
  " C + A *[M], [M] -> B",   // LAMC
  " C + B *[M]",             // BMC
  " C +[M]*[M], [M] -> B",   // LMMC
  " C +[M]*[M]",             // MMC
  " C +[M]*[M],   A -> B",   // XMMC
  " C +[M]",                 // 1MC
  " C +[M]    , [M] -> B",   // L1MC
  " C + A * A",              // AAC
  " C + A * A , [M] -> B",   // LAAC
  " C + B * A",              // BAC
  " B + C * A",              // CAB
  " D + C * A",              // CAD
  "     C * A , [M] -> B",   // LCA
  "[M]+ C * A",              // CAM
  " C + D * A",              // DAC
  " C + B * B",              // BBC
  "     C * B",              // CB
  "[M]+ C * B",              // CBM
  " C + A     ,   A -> B",   // X1AC
  " C + A",                  // 1AC
  " C + A     , [M] -> B",   // L1AC
  "LOG16(B)",                // LOGB
  "16^B",                    // EXPB
  "     C * A ,   A -> [M]", // SCA
  "     C * A ,   A -> [M], A -> B", // SXCA
  "     C * B ,   A -> [M]", // SCB
  " A + B * U ,   A -> [M]", // SBUA
  " A + C * U ,   A -> [M]", // SCUA
  " B + C * U ,   A -> [M]", // SCUB
  "     C * U ,   A -> [M]", // SCU
  " C + B * U ,   A -> [M]", // SBUC

  " C + A * A ,   A -> [M]", // SAAC
  " C + A * A ,   A -> [M], A -> B", // SXAAC
  " C + B * A ,   A -> [M]", // SBAC
  " C + B * B ,   A -> [M]", // SBBC
  " A + C * B ,   A -> [M]", // SCBA
  " B + C * A ,   A -> [M]", // SCAB
  "LOG16(B)   ,   A -> [M]", // SLOGB
  "16^B       ,   A -> [M]", // SEXPB
  " C + A * U ,   A -> [M]", // SAUC
  " C + A     ,   A -> [M]", // S1AC
  " C + A     ,   A -> [M], A -> B", // SX1AC
  " C + B     ,   A -> [M]", // S1BC
  " A + 0 * U ,   A -> [C&B + M]",   // SINDF
  " A + 0 * U ,   A -> [C&B + M]",   // SINDI
  " 0 -> U, B&A",   // ANDB
  " 0 -> U, C&A",   // ANDC
  "INTEGER : A + B",   // ADDB
  "INTEGER : A - B",   // SUBB
  "INTEGER : A * B",   // MLTB
  "if C, C -> A, else 0 -> A",   // C
  "if C, treat next C instruction as NOPs",   // SKIP
  " C + B",            // 1BC
  "[C&B + M]",   // LINDF
  "[C&B + M]"    // LINDI
};  

/*
** Check if the object object filename has good extension  and exists.
// Returns 0 if success, 1 otherwise
*/
int CheckObjFilename(char* fn){
  int len = strlen(fn);
  
  // check if '.obj' is present at end of fn. Add '.obj' if it is not the case
  if ((len<5) || (fn[len-1]!='j') || (fn[len-2]!='b') 
      || (fn[len-3]!='o') || (fn[len-4]!='.')){
    strcat(fn, ".obj");
    len+=4;
  }

  // check if the file exist
  FILE *flux=fopen(fn, "r");
  if (flux==NULL){
    return 1;
  }
  fclose(flux);

  return 0;
}

/* check if the object line is valid :
**   1. has 10 caracters
**   2. is only composed of hexadecimal caracters
** Return 0 if success, i if step i has failed.
*/
int CheckLine(char *CmdLine, int len){
  int i;

  // removing EOL caracter(s)
  while ((CmdLine[len-1]==0x0A) || (CmdLine[len-1]==0x0D)){
    CmdLine[len-1]='\0';
    len--;
  }
  if (len!=10)
    return 1;

  // check if all caracters are valid.
  for (i=0;i<10;i++){
    if ((CmdLine[i]<'0')
	|| ((CmdLine[i]>'9') && (CmdLine[i]<'A'))
	|| ((CmdLine[i]>'F') && (CmdLine[i]<'a'))
	|| (CmdLine[i]>'f')){
      return 2;
    }
  }
  
  return 0;
}

/* convert the object line string
** in two integer : wParam (bits 33 to 40)
** and lParam (bits 1 to 32)
** Returns 0 if success, 1 if failure
*/
int ConvertToInt(char *Cmd, UINT32 *wparam, UINT32 *lparam){
  int i;
  UINT32 val=0;

  // initialize parameters
  (*wparam)=(*lparam)=0;

  // browse the string
  for (i=0;i<10;i++){
    switch (Cmd[i]){
    case '0': val = 0; break;    case '1': val = 1; break;
    case '2': val = 2; break;    case '3': val = 3; break;
    case '4': val = 4; break;    case '5': val = 5; break;
    case '6': val = 6; break;    case '7': val = 7; break;
    case '8': val = 8; break;    case '9': val = 9; break;
    case 'A': case 'a': val = 10; break;
    case 'B': case 'b': val = 11; break;
    case 'C': case 'c': val = 12; break;
    case 'D': case 'd': val = 13; break;
    case 'E': case 'e': val = 14; break;
    case 'F': case 'f': val = 15; break;
    }
    // only the two caracters ont the left (i=0,1) contributes to wParam,
    // the other 8 ones contributes to lParam.
    if (i>1)   
      (*lparam)+=(val<<(4*(9-i)));
    else
      (*wparam)+=(val<<(4*(1-i)));
  }

  return 0;
}

/* 
** Extract the OpCode out of wParam
** return 0 if succeeded, 1 otherwise
** if the function succeed, the index of the 
** opcode in MnemonicTable is returned through 'op_index'
*/
int GetOPCode(UINT32 wparam, int *op_index){
  int i=0;
  INT32 op;

  // retreive OpCode
  op = wparam>>2;

  // browse the MnemonicTable
  do{
    if (MnemonicTable[i].OpCode==op){
      (*op_index) = i;
      return 0;
    }      
    i++;
  } while (i<NB_OPCODES);

  return 1;
} 

/*
** This function disassemble the code found in the file 'ObjectFileName' 
** if it exist. 
** It fills the structure AL3101_disasm (see def of AL3101_disasm_struct type).
** It return DISASM_WRONG_FILENAME if an error occured while opening 
**                                 or reading the object file,
**           DISASM_ERRORS         if errors were encountered.
**           the number of codes   if disassemble successfully
*/
int Disassemble(char *ObjectFileName, char *DisasmFileName, 
		AL3101_Disasm_struct *AL3101_disasm){

  FILE *flux_object;        // stream for the object file
  FILE *flux_disasm = NULL; // stream for the output file (if exists)
  char Obj_fn[255];         // buffer to copy the object file
  
  int FileSize;             // Size of the input file
  int ErrorNumber=0;        // Number of detected errors

  char CmdLine[255];        // command line in the input file
  int LineIndex=0;          // index of the current line in the file 
  int len;                  // length of the Command line
  int opcode_index;         // index of the opcode in the Mnemonictable
  UINT32 wParam, lParam;    // code for assembled parameters and instruction
  char CommandStr[255];     // string with the disassembled command
  
  int eb;
  char tmp[80];

  // Copy the filename to a local variable.
  // This is useful when .obj has to be added. In such a case we have no
  // interference with the calling application.
  strcpy(Obj_fn, ObjectFileName);

  // check if it is a valid filename
  if (CheckObjFilename(Obj_fn)!=0){
    ERROR_1("File '%s' doesn't exists. Giving up.\n", ObjectFileName);
    return DISASM_FILE_NOT_EXISTS;
  }
  flux_object = fopen(Obj_fn,  "r");

  // if asked for the disassembled code, create the file.
  // We don't check if the file has extension .asm nor if it already exists.
  // In the latter case, the file is simply overwritten.
  if (DisasmFileName!=NULL){
    flux_disasm = fopen(DisasmFileName, "w+");
    fprintf(flux_disasm, "; Generated by AL3101 Disassembler\n");
  }

  // enter into the main disassemble loop
  sprintf(CmdLine, "");
    
  // calculating size of the input file
  fseek(flux_object, 0L, SEEK_END);
  FileSize=ftell(flux_object);
  fseek(flux_object, 0L, SEEK_SET);

  // browse command lines until EOF reached or an error occured
  while ((LineIndex<1024) && (fgets(CmdLine, MAX_CMDLINE_SIZE, flux_object)!=NULL)){

    LineIndex++;
   
    // check the syntax of the line
    len = strlen(CmdLine);
    switch(CheckLine(CmdLine, len)){
    case 1:    // the object line has not correct length
      ErrorNumber++;
      ERROR_1("Disasm:%d: Line has not a correct length\n", LineIndex);
      continue;
      break;
    case 2:    // object line contains invalid caracters
      ErrorNumber++;
      ERROR_1("Disasm:%d: Line contains non-valid caracters\n", LineIndex);
      continue;
      break;      
    }
    
    ConvertToInt(CmdLine, &wParam, &lParam);

    // store binary code in simulator structure
    AL3101_disasm[LineIndex-1].lObj = lParam;
    AL3101_disasm[LineIndex-1].wObj = wParam;

    // get Mnemonic and get the associated opcode
    if (GetOPCode(wParam, &opcode_index)!=0){
      ErrorNumber++;
      ERROR_2("%d: Unrecognized opcode '%x'\n", LineIndex, (wParam>>2));
      continue;
    }
    AL3101_disasm[LineIndex-1].op=MnemonicTable[opcode_index].OpCode;

    // remove OpCode from wParam
    wParam%=0x4;

    // call the disassembling function in order to fill the parameter fields
    // of the AL3101_disasm structure
    // A String containing parameters is returned through tmp.
    eb=MnemonicTable[opcode_index].f_disasm(wParam, lParam, 
					    &(AL3101_disasm[LineIndex-1]), tmp);
    /* if eb=1, we have an extra-bit = 1, this doesn't change the opcode 
    ** (neither the disasm function to call !!!) but the line index in the 
    ** Table has to be increased by 1.
    */

    // Take into account the offset for output addresses (for the simulator)
    SimulatorOutputOffset(AL3101_disasm[LineIndex-1].op, &(AL3101_disasm[LineIndex-1].m));

    // fills the remaining fields
    AL3101_disasm[LineIndex-1].s_index=opcode_index;
    if (eb!=0) AL3101_disasm[LineIndex-1].s_index++;
    AL3101_disasm[LineIndex-1].ln=LineIndex;

    // save disassembled string in structure
    sprintf(CommandStr, "%s\t%s ; %s", 
	    MnemonicTable[opcode_index].Mnemonic, 
	    tmp, Comments[opcode_index]);
    strcpy(AL3101_disasm[LineIndex-1].CmdStr, CommandStr);
    // if user asked for a disassemble file add command line to it.
    if (DisasmFileName!=NULL){
      while (strlen(tmp)<21) strcat(tmp, " ");
      fprintf(flux_disasm, "%s\n", CommandStr);
      AL3101_disasm[LineIndex-1].ln++;   // take the first line into account
    }

    sprintf(CmdLine, "");
  };

  // if we haven't read 'Filesize' bytes, an error occured
  if (ftell(flux_object)<FileSize){
    fclose(flux_object);
    ERROR("Error before EOF\n\n");
    return DISASM_READ_ERROR;
  }

  // close the disassembled file (if asked)
  if (DisasmFileName!=NULL)
    fclose(flux_disasm);

  // close file, show number of errors
  fclose(flux_object);
  if (ErrorNumber!=0)
    ERROR_1("Disassembling found %d Errors\n\n", ErrorNumber);    
  else
    ERROR("Disassembling successful\n\n");
    

  return LineIndex;
}
