// FileIO_WAV.cpp: implementation of the WAV File functions.
//
//////////////////////////////////////////////////////////////////////

#include "FileIO_WAV.h"
#include "FileIO_WAV_internal.h"

#include <stdlib.h>		// for the malloc function
#include <string.h>		// for the string routines in error functions


//////////////////////////////////////////////////////////////////////
// init & clean functions
//////////////////////////////////////////////////////////////////////
INT32 FileIO_WAV_Init(void **pInstance, CHAR *FileName, INT32 Mode, FileIO_SndCfg *cfg, INT32 *LeftSampleMemory, INT32 *RightSampleMemory)
{
	INT32 errorcode;
	FileIO_WAV_Struct *pInst;

	// Allocate the space for the instance structure
	pInst = new FileIO_WAV_Struct;
	if (pInst == NULL) return FILEIO_CANNOT_ALLOCATE_INSTANCE;
	*pInstance = (void *)pInst;

	// Instantiate the CWaveFile class
	pInst->File = new CWaveFile;
	if (pInst->File == NULL)
	{
		delete pInst;
		return FILEIO_CANNOT_ALLOCATE_INSTANCE;
	}

	pInst->Mode = Mode;

	if (pInst->Mode != 0)
	{
		errorcode = pInst->File->OpenForWrite(FileName, cfg->SamplingFrequency, cfg->NumberOfBytesInSample, cfg->NumberOfChannels);
		// test if errorcode is good
		if (errorcode != FILEIO_NOERROR)
		{
			FileIO_WAV_Clean(pInst);
			return errorcode;
		}
	}
	else
	{
		errorcode = pInst->File->OpenForRead(FileName);
		// test if errorcode is good
		if (errorcode != FILEIO_NOERROR)
		{
			FileIO_WAV_Clean(pInst);
			return errorcode;
		}
		// Write the Format in the FileIO_SndCfg structure
		cfg->SamplingFrequency = pInst->File->GetSamplingRate();
		cfg->NumberOfBytesInSample = pInst->File->GetBytesPerSample();
		cfg->NumberOfChannels = pInst->File->GetNumChannels();
		// Check if The number of channel wanted is not OK with what available in the file
		if (((RightSampleMemory == NULL) && (cfg->NumberOfChannels == 2)) || ((RightSampleMemory != NULL) && (cfg->NumberOfChannels != 2)))
		{
			FileIO_WAV_Clean(pInst);
			return FILEIO_WRONG_CHANNEL_NUMBER;
		}
	}

	pInst->NumberOfChannels = cfg->NumberOfChannels;		// Number of channels contained in the file
	pInst->SamplingFrequency = cfg->SamplingFrequency;		// The sampling frequency of the file
	pInst->File->Looping = 0;
	
	// do the DSP memory related stuff initialization
	pInst->LeftData = LeftSampleMemory;		// Pointer to the left sample memory
	pInst->RightData = RightSampleMemory;	// Pointer to the right sample memory
	
	return FILEIO_NOERROR;	// no problem occurs
}

INT32 FileIO_WAV_Clean(void *pInstance)
{
	INT32 errorcode;
	FileIO_WAV_Struct *pInst = (FileIO_WAV_Struct *)pInstance;

	// "closing" the class
	errorcode = pInst->File->Close();
	if (errorcode != FILEIO_NOERROR) return errorcode;

	// freeing the class instance
	delete pInst->File;

	// freeing the instance structure
	if (pInst != NULL) delete pInst;

	return FILEIO_NOERROR;
}


//////////////////////////////////////////////////////////////////////
// "processing" functions
//////////////////////////////////////////////////////////////////////
INT32 FileIO_WAV_ReadSample(void *pInstance)
{
	FileIO_WAV_Struct *pInst = (FileIO_WAV_Struct *)pInstance;
	INT16 Buffer[2];
	INT32 tempint;

	// Check if we try to read in a file which has been opened in that mode
	if (pInst->Mode != 0)
	{
		return FILEIO_WRONG_FILE_MODE;				// Try to read in a file that is in write mode
	}
	
	
	// Check if data is still available in the buffer
	INT32 result = pInst->File->ReadData(Buffer, pInst->NumberOfChannels);
	if (result != FILEIO_NOERROR) return result;

	// let's now transfer data from the buffer to the memory
	tempint = Buffer[0];							// first copy the 16 bits value in a 32 bits one
	tempint = tempint << 8;							// then prepare the data to fit a S3.24 format
	*(pInst->LeftData) = tempint;					// copy left/mono channel to memory
	if (pInst->NumberOfChannels == 2)				// check if there's a second channel
	{
		tempint = Buffer[1];						// first copy the 16 bits value in a 32 bits one
		tempint = tempint << 8;						// then prepare the data to fit a S3.24 format
		*(pInst->RightData) = tempint;				// copy right channel to memory
	}
	
	return FILEIO_NOERROR;	// no problem occurs
}

	
INT32 FileIO_WAV_WriteSample(void *pInstance)
{
	FileIO_WAV_Struct *pInst = (FileIO_WAV_Struct *)pInstance;
	INT16 Buffer[2];
	INT32 tempint;

	// Check if we try to read in a file which has been opened in that mode
	if (pInst->Mode == 0)
	{
		return FILEIO_WRONG_FILE_MODE;				// Try to write in a file that is in read mode
	}

	// let's transfer data from the memory to the buffer
	tempint = *(pInst->LeftData);
	if (tempint > (INT32)0x007FFFFF) tempint = 0x007FFFFF;
	if (tempint < (INT32)0xFF800000) tempint = 0xFF800000;
	tempint = tempint >> 8;
	Buffer[0] = (INT16)tempint;							// copy left/mono channel to buffer
	if (pInst->NumberOfChannels == 2)					// check if there's a second channel
	{
		tempint = *(pInst->RightData);
		if (tempint > (INT32)0x007FFFFF) tempint = 0x007FFFFF;
		if (tempint < (INT32)0xFF800000) tempint = 0xFF800000;
		tempint = tempint >> 8;
		Buffer[1] = (INT16)tempint;						// copy right channel to buffer
	}

	return pInst->File->WriteData(Buffer, pInst->NumberOfChannels);
}


//////////////////////////////////////////////////////////////////////
// file function(s)
//////////////////////////////////////////////////////////////////////
INT32 FileIO_WAV_GetPosition(void *pInstance, INT32 *CurrentPosition)
{
	FileIO_WAV_Struct *pInst = (FileIO_WAV_Struct *)pInstance;

	// use standard function to get position and return if an error occured
	INT32 size = pInst->File->GetCurrentFilePosition() - 2 * sizeof(RiffChunkHeader) - sizeof(WaveFormat_Chunk) - 2;
	if (size == -1)
	{
		*CurrentPosition = 0;
		return FILEIO_CANNOT_GET_POSITION;
	}
	else
	{
		if (pInst->NumberOfChannels == 1) *CurrentPosition = size / 2;
		else *CurrentPosition = size / 4;
		return FILEIO_NOERROR;
	}
}

INT32 FileIO_WAV_GetFileLength(void *pInstance, INT32 *FileSize)
{
	INT32 size;

	FileIO_WAV_Struct *pInst = (FileIO_WAV_Struct *)pInstance;

	// Get the number of samples contained in the file
	size = pInst->File->GetNumSamples();

	*FileSize = size;

	return FILEIO_NOERROR;
}



//////////////////////////////////////////////////////////////////////
// control function(s)
//////////////////////////////////////////////////////////////////////
INT32 FileIO_WAV_SetLoopingMode(void *pInstance, INT32 Looping)
{
	FileIO_WAV_Struct *pInst = (FileIO_WAV_Struct *)pInstance;
	pInst->File->Looping = Looping;

	return FILEIO_NOERROR;
}



//////////////////////////////////////////////////////////////////////
// "text" function(s)
//////////////////////////////////////////////////////////////////////
void FileIO_WAV_GetLastErrorString(CHAR *str, INT32 LastErrorCode, INT32 MaxNumberOfCharacters)
{
	if ((LastErrorCode & FILEIO_BASE_ERR) == FILEIO_BASE_ERR)		// check if the error is a FileIO error
	{
		if ((LastErrorCode & FILEIO_GEN_BASE_ERR) == FILEIO_GEN_BASE_ERR)
		{
			FileIO_GetLastErrorString(str, LastErrorCode, MaxNumberOfCharacters);
			return;
		}
		else if ((LastErrorCode & FILEIO_WAV_BASE_ERR) == FILEIO_WAV_BASE_ERR)
		{
			// do nothing here, wait for the switch below
		}
		else
		{
			strncpy(str, "Error type not known", MaxNumberOfCharacters - 1);
		}
	}
	else															// this error is not a FileIO error
	{
		strncpy(str, "Not a FileIO error", MaxNumberOfCharacters - 1);
		return;
	}


	switch(LastErrorCode)
	{
		case FILEIO_WAV_NOERROR:
			strncpy(str, "WAV: No error occurred", MaxNumberOfCharacters - 1);
			break;
		default:
			strncpy(str, "WAV: Error not known", MaxNumberOfCharacters - 1);
			break;
	}
	return;
}



void FileIO_WAV_GetHelp(CHAR *str, INT32 LineNumber, INT32 MaxNumberOfCharacters)
{
	switch(LineNumber)
	{
		case 0:
			strncpy(str, "WAV files are read by these functions.", MaxNumberOfCharacters - 1);
			break;
		case 1:
			strncpy(str, "WAV file format used is the standard audio file format", MaxNumberOfCharacters - 1);
			break;
		case 2:
			strncpy(str, "There's no support for 8 bits files and for more than 4GB files.", MaxNumberOfCharacters - 1);
			break;
		default:
			strncpy(str, "WAV file functions", MaxNumberOfCharacters - 1);
	}
}