// FileIO_WAV_Wave.cpp: implementation of the CWaveFile class.
//
//////////////////////////////////////////////////////////////////////

#include "FileIO_WAV_Wave.h"

// Standard includes
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef WIN32
#include <unistd.h>
#endif

#include "./../FileIO_Generic.h"


//////////////////////////////////////////////////////////////////////
// Helper function
//////////////////////////////////////////////////////////////////////
UINT32 FourCC ( CHAR *ChunkName )
{
   UINT32 retbuf = 0x20202020;   // four spaces (padding)
   CHAR *p = ((CHAR *)&retbuf);

   // Remember, this is Intel format!
   // The first character goes in the LSB
   for ( INT32 i=0; i<4 && ChunkName[i]; i++ )
   {
	  *p++ = ChunkName[i];
   }
   return retbuf;
}


INT32 GetFileLength(FILE *file, INT32 *FileSize)
{
	INT32 result;
	INT32 CurrentPosition, size;

	// Get the current position in the file and return if impossible
	CurrentPosition = ftell(file);
	if (CurrentPosition == -1) return FILEIO_CANNOT_GET_POSITION;

	
	// Seek to the end of file
	result = fseek(file, 0, SEEK_END);
	if (result != 0) return FILEIO_PROBLEM_WHILE_SEEKING;
	
	// Get the position of the end of file
	size = ftell(file);
	if (size == -1) return FILEIO_CANNOT_GET_POSITION;


	//  Seek to the position where the file was before
	result = fseek(file, CurrentPosition, SEEK_SET);
	if (result != 0) return FILEIO_PROBLEM_WHILE_SEEKING;

	*FileSize = size;

	return FILEIO_NOERROR;
}


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CWaveFile::CWaveFile()
{
	file = 0;
	Mode = 0;
	num_samples = 0;
	riff_header.ckID = FourCC("RIFF");
	riff_header.ckSize = 0;
	pcm_data.ckID = FourCC("data");
	pcm_data.ckSize = 0;
}

CWaveFile::~CWaveFile()
{
	if ( file != NULL )
	{
	  Close();
	}
}

INT32 CWaveFile::OpenForRead ( CHAR *Filename )
{
	INT32 retcode = FILEIO_NOERROR;
	INT32 FileSize;

	file = fopen ( Filename, "rb" );
	if (file != NULL)
	{
		// Try to read the RIFF header...

		if ( fread ( &riff_header, sizeof(riff_header), 1, file ) != 1 )
		{
			fclose(file);
			file = NULL;
		}
		else
		{
			Mode = 0;
		}
	}
	else
	{
		retcode = FILEIO_CANNOT_OPEN_FILE;
	}

	if ( retcode == FILEIO_NOERROR )
	{
		retcode = FILEIO_NOERROR;
		CHAR Ref[5] = {'W','A','V','E',0};
		

		for (INT32 NumBytes = 0; NumBytes < 4; NumBytes++)
		{
			if (fgetc(file) != Ref[NumBytes]) retcode = FILEIO_READ_ERROR;
		}


		if ( retcode == FILEIO_NOERROR )
		{
			retcode = ReadFunction ( &wave_format, sizeof(wave_format) );
	
			if ( retcode == FILEIO_NOERROR &&
				!wave_format.VerifyValidity() )
			{
				// This isn't standard PCM, so we don't know what it is!

				retcode = FILEIO_TYPE_NOT_SUPPORTED;
			}

			if (wave_format.data.nBitsPerSample != 16)		// we do not care about 8 bits files
				retcode = FILEIO_TYPE_NOT_SUPPORTED;

			if ( retcode == FILEIO_NOERROR )
			{
				pcm_data_offset = ftell(file);

				// Figure out number of samples from
				// file size, current file position, and
				// WAVE header.

				retcode = ReadFunction ( &pcm_data, sizeof(pcm_data) );
				GetFileLength(file, &FileSize);
				num_samples = FileSize - GetCurrentFilePosition();
				num_samples /= GetNumChannels();
				num_samples /= GetBytesPerSample();
			}
		}
	}

	return retcode;
}


INT32 CWaveFile::OpenForWrite ( CHAR	*Filename,
								UINT32	SamplingRate,
								UINT32	BytesPerSample,
								UINT32	NumChannels )
{
	// Verify parameters...

	if ( !Filename ||
			BytesPerSample != 2 ||
			NumChannels < 1 || NumChannels > 2 )
	{
		return FILEIO_TYPE_NOT_SUPPORTED;
	}

	wave_format.data.Config ( SamplingRate, BytesPerSample * 8, NumChannels );

	INT32 retcode = FILEIO_NOERROR;
	file = fopen ( Filename, "wb" );
	if (file != NULL)
	{
		// Write the RIFF header...
		// We will have to come back later and patch it!

		if ( fwrite ( &riff_header, sizeof(riff_header), 1, file ) != 1 )
		{
			fclose(file);
			unlink(Filename);
			file = NULL;
		}
		else
		{
			Mode = 1;
		}
	}
	else
	{
		retcode = FILEIO_CANNOT_OPEN_FILE;
	}

	if ( retcode == FILEIO_NOERROR )
	{
		retcode = WriteFunction((void *)"WAVE", 4, &riff_header);

		if ( retcode == FILEIO_NOERROR )
		{
			retcode = WriteFunction(&wave_format, sizeof(wave_format), &riff_header);

			if ( retcode == FILEIO_NOERROR )
			{
				pcm_data_offset = GetCurrentFilePosition();
				retcode = WriteFunction ( &pcm_data, sizeof(pcm_data), &riff_header);
			}
		}
	}

	return retcode;
}


INT32 CWaveFile::Close()
{
	INT32 retcode = FILEIO_NOERROR;
   
	// Test if instance is initialized
	if ( !file ) return FILEIO_INSTANCE_NOT_INITIALIZED;

	if ( Mode != 0 )
	{
		if ( fflush(file) ||
				fseek ( file, pcm_data_offset, SEEK_SET ) )
		{
			retcode = FILEIO_PROBLEM_WHILE_SEEKING;
		}
		else
		{
			retcode = WriteFunction ( &pcm_data, sizeof(pcm_data), &riff_header );
		}
	}
    

	if ( retcode == FILEIO_NOERROR )
	{
		if (Mode != 0)
		{
			if ( fflush(file) ||
					fseek(file,0,SEEK_SET) ||
					fwrite ( &riff_header, sizeof(riff_header), 1, file ) != 1 ||
					fclose(file) )
			{
				retcode = FILEIO_CANNOT_CLOSE_FILE;
			}
		}
		else
		{
			fclose(file);
		}

		file = NULL;
	}

	return retcode;
}


INT32 CWaveFile::SeekToSample ( UINT32 SampleIndex )
{
	if ( SampleIndex >= GetNumSamples() )
	{
		return FILEIO_TYPE_NOT_SUPPORTED;
	}

	UINT32 SampleSize = (GetBytesPerSample() * 8 + 7) / 8;

	INT32 rc = FILEIO_NOERROR;
	INT32 offset;
   
	offset = pcm_data_offset + sizeof(pcm_data) + SampleSize * GetNumChannels() * SampleIndex;
	
	fflush(file);

	if ( fseek(file, offset, SEEK_SET) )
	{
		rc = FILEIO_PROBLEM_WHILE_SEEKING;
	}
	else
	{
		rc = FILEIO_NOERROR;
	}

	return rc;
}


UINT32 CWaveFile::GetSamplingRate()
{
   return wave_format.data.nSamplesPerSec;
}


UINT32 CWaveFile::GetBytesPerSample()
{
   return wave_format.data.nBitsPerSample / 8;
}


UINT32 CWaveFile::GetNumChannels()
{
   return wave_format.data.nChannels;
}


UINT32 CWaveFile::GetNumSamples()
{
   return num_samples;
}


INT32 CWaveFile::WriteData ( INT16 *data, UINT32 numData )
{
	UINT32 extraBytes = numData * sizeof(INT16);
	pcm_data.ckSize += extraBytes;
	return WriteFunction ( data, extraBytes, &riff_header );
}


INT32 CWaveFile::ReadData ( INT16 *data, UINT32 numData )
{
	return ReadFunction ( data, numData * sizeof(INT16) );
}


INT32 CWaveFile::WriteFunction ( void *Data, UINT32 NumBytes, RiffChunkHeader *header)
{
	if (file == NULL) return FILEIO_INSTANCE_NOT_INITIALIZED;

	if ( Mode == 0 )
	{
		return FILEIO_WRONG_FILE_MODE;
	}

	if ( fwrite ( Data, NumBytes, 1, file ) != 1 )
	{
		return FILEIO_WRITE_ERROR;
	}

	header->ckSize += NumBytes;

	return FILEIO_NOERROR;
}

INT32 CWaveFile::ReadFunction ( void *Data, UINT32 NumBytes )
{
	if (file == NULL) return FILEIO_INSTANCE_NOT_INITIALIZED;

	if ( fread(Data,NumBytes,1,file) != 1 )
	{
		if (!feof(file)) return FILEIO_READ_ERROR;
		else
		{
			if (Looping)
			{
				memset(Data, 0, NumBytes * sizeof(INT8));
				return SeekToSample(0);
			}
			else return FILEIO_EOF_REACHED;
		}
	}

	return FILEIO_NOERROR;
}
/*--- end of file FileIO_WAV_Wave.cpp ---*/